@@ -357,7 +357,23 @@ MODULE_SCOPE Tcl_Obj *tclpkcs11_bytearray_to_string(const unsigned char *data, u
 		buf[bufidx++] = alphabet[data[idx] & 0xf];
 	}
 
-	retval = Tcl_NewByteArrayObj((unsigned char *) buf, bufidx);
+	retval = Tcl_NewStringObj(buf, bufidx);
+
+	return(retval);
+}
+
+MODULE_SCOPE Tcl_Obj *tclpkcs11_bytearray_to_string_from_obj(Tcl_Obj *data) {
+	unsigned char *buf;
+	int buflen;
+	Tcl_Obj *retval;
+
+	if (data == NULL) {
+		return(Tcl_NewObj());
+	}
+
+	buf = Tcl_GetByteArrayFromObj(data, &buflen);
+
+	retval = tclpkcs11_bytearray_to_string(buf, buflen);
 
 	return(retval);
 }
@@ -680,11 +696,15 @@ MODULE_SCOPE int tclpkcs11_load_module(ClientData cd, Tcl_Interp *interp, int ob
 
 	chk_rv = pkcs11_function_list->C_Initialize(&initargs);
 	if (chk_rv == CKR_CANT_LOCK) {
+		/*
+		 * If the PKCS#11 module rejects using our locking
+		 * functions, retry with OS locking functions.
+		 */
 		initargs.CreateMutex = NULL;
 		initargs.DestroyMutex = NULL;
 		initargs.LockMutex = NULL;
 		initargs.UnlockMutex = NULL;
-		initargs.flags = 0;
+		initargs.flags = CKF_OS_LOCKING_OK;
 		initargs.LibraryFlags = NULL;
 		initargs.pReserved = NULL;
 
@@ -998,7 +1018,7 @@ MODULE_SCOPE int tclpkcs11_list_certs(ClientData cd, Tcl_Interp *interp, int obj
 	                           {CKA_VALUE, NULL, 0}
 	}, *curr_attr;
 	CK_ULONG curr_attr_idx;
-	CK_OBJECT_CLASS *objectclass;
+	CK_OBJECT_CLASS objectclass, *objectclass_p;
 	CK_RV chk_rv;
 
 	if (!cd) {
@@ -1053,6 +1073,255 @@ MODULE_SCOPE int tclpkcs11_list_certs(ClientData cd, Tcl_Interp *interp, int obj
 		return(TCL_ERROR);
 	}
 
+	ret_list = Tcl_NewObj();
+	while (1) {
+		chk_rv = handle->pkcs11->C_FindObjects(handle->session, &hObject, 1, &ulObjectCount);
+		if (chk_rv != CKR_OK) {
+			Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
+
+			handle->pkcs11->C_FindObjectsFinal(handle->session);
+
+			return(TCL_ERROR);
+		}
+
+		if (ulObjectCount == 0) {
+			break;
+		}
+
+		if (ulObjectCount != 1) {
+			Tcl_SetObjResult(interp, Tcl_NewStringObj("FindObjects() returned a weird number of objects.", -1));
+
+			handle->pkcs11->C_FindObjectsFinal(handle->session);
+
+			return(TCL_ERROR);
+		}
+
+		for (curr_attr_idx = 0; curr_attr_idx < (sizeof(template) / sizeof(template[0])); curr_attr_idx++) {
+			curr_attr = &template[curr_attr_idx];
+			if (curr_attr->pValue) {
+				ckfree(curr_attr->pValue);
+			}
+
+			curr_attr->pValue = NULL;
+			curr_attr->ulValueLen = 0;
+		}
+
+		/* Determine size of values to allocate */
+		chk_rv = handle->pkcs11->C_GetAttributeValue(handle->session, hObject, template, sizeof(template) / sizeof(template[0]));
+		if (chk_rv == CKR_ATTRIBUTE_TYPE_INVALID || chk_rv == CKR_ATTRIBUTE_SENSITIVE || chk_rv == CKR_BUFFER_TOO_SMALL) {
+			chk_rv = CKR_OK;
+		}
+		if (chk_rv != CKR_OK) {
+			/* Skip this object if we are not able to process it */
+			continue;
+		}
+
+		/* Allocate values */
+		for (curr_attr_idx = 0; curr_attr_idx < (sizeof(template) / sizeof(template[0])); curr_attr_idx++) {
+			curr_attr = &template[curr_attr_idx];
+
+			if (((CK_LONG) curr_attr->ulValueLen) != ((CK_LONG) -1)) {
+				curr_attr->pValue = (void *) ckalloc(curr_attr->ulValueLen);
+			}
+		}
+
+		/* Populate template values */
+		chk_rv = handle->pkcs11->C_GetAttributeValue(handle->session, hObject, template, sizeof(template) / sizeof(template[0]));
+		if (chk_rv != CKR_OK && chk_rv != CKR_ATTRIBUTE_SENSITIVE && chk_rv != CKR_ATTRIBUTE_TYPE_INVALID && chk_rv != CKR_BUFFER_TOO_SMALL) {
+			/* Return an error if we are unable to process this entry due to unexpected errors */
+			for (curr_attr_idx = 0; curr_attr_idx < (sizeof(template) / sizeof(template[0])); curr_attr_idx++) {
+				curr_attr = &template[curr_attr_idx];
+				if (curr_attr->pValue) {
+					ckfree(curr_attr->pValue);
+				}
+			}
+
+			Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
+
+			handle->pkcs11->C_FindObjectsFinal(handle->session);
+
+			return(TCL_ERROR);
+		}
+
+		/* Extract certificate data */
+		obj_label = NULL;
+		obj_id = NULL;
+		obj_cert = NULL;
+		objectclass_p = NULL;
+		for (curr_attr_idx = 0; curr_attr_idx < (sizeof(template) / sizeof(template[0])); curr_attr_idx++) {
+			curr_attr = &template[curr_attr_idx];
+
+			if (!curr_attr->pValue) {
+				continue;
+			}
+
+			if (curr_attr->type != CKA_CLASS && !objectclass_p) {
+				ckfree(curr_attr->pValue);
+				curr_attr->pValue = NULL;
+
+				continue;
+			}
+
+			switch (curr_attr->type) {
+				case CKA_CLASS:
+					objectclass_p = (CK_OBJECT_CLASS *) curr_attr->pValue;
+					objectclass = *objectclass_p;
+					break;
+				case CKA_LABEL:
+					obj_label = Tcl_NewStringObj(curr_attr->pValue, curr_attr->ulValueLen);
+					break;
+				case CKA_ID:
+					/* Convert the ID into a readable string */
+					obj_id = tclpkcs11_bytearray_to_string(curr_attr->pValue, curr_attr->ulValueLen);
+
+					break;
+				case CKA_VALUE:
+					obj_cert = Tcl_NewByteArrayObj(curr_attr->pValue, curr_attr->ulValueLen);
+
+					break;
+			}
+
+			ckfree(curr_attr->pValue);
+			curr_attr->pValue = NULL;
+		}
+
+		/* Add this certificate data to return list, if all found */
+		if (obj_label == NULL || obj_id == NULL || obj_cert == NULL) {
+			continue;
+		}
+
+		if (objectclass != CKO_CERTIFICATE) {
+			continue;
+		}
+
+		/* Create the current item list */
+		curr_item_list = Tcl_NewObj();
+		Tcl_ListObjAppendElement(interp, curr_item_list, Tcl_NewStringObj("pkcs11_handle", -1));
+		Tcl_ListObjAppendElement(interp, curr_item_list, tcl_handle);
+
+		Tcl_ListObjAppendElement(interp, curr_item_list, Tcl_NewStringObj("pkcs11_slotid", -1));
+		Tcl_ListObjAppendElement(interp, curr_item_list, tcl_slotid);
+
+		Tcl_ListObjAppendElement(interp, curr_item_list, Tcl_NewStringObj("pkcs11_id", -1));
+		Tcl_ListObjAppendElement(interp, curr_item_list, obj_id);
+
+		Tcl_ListObjAppendElement(interp, curr_item_list, Tcl_NewStringObj("pkcs11_label", -1));
+		Tcl_ListObjAppendElement(interp, curr_item_list, obj_label);
+
+		/* Call "::pki::x509::parse_cert" to parse the cert */
+		parse_cert_cmd = Tcl_NewObj();
+		Tcl_ListObjAppendElement(interp, parse_cert_cmd, Tcl_NewStringObj("::pki::x509::parse_cert", -1));
+		Tcl_ListObjAppendElement(interp, parse_cert_cmd, obj_cert);
+
+		tcl_rv = Tcl_EvalObjEx(interp, parse_cert_cmd, 0);
+		if (tcl_rv != TCL_OK) {
+			continue;
+		}
+
+		/* Add results of [parse_cert] to our return value */
+		Tcl_ListObjAppendList(interp, curr_item_list, Tcl_GetObjResult(interp));
+
+		/*
+		 * Override the "type" so that [array set] returns our new
+		 * type, but we can still parse through the list and figure
+		 * out the real subordinate type
+		 */
+		Tcl_ListObjAppendElement(interp, curr_item_list, Tcl_NewStringObj("type", -1));
+		Tcl_ListObjAppendElement(interp, curr_item_list, Tcl_NewStringObj("pkcs11", -1));
+
+		/* Add the current item to the return value list */
+		Tcl_ListObjAppendElement(interp, ret_list, curr_item_list);
+	}
+
+	/* Terminate search */
+	handle->pkcs11->C_FindObjectsFinal(handle->session);
+
+	/* Return */
+	Tcl_SetObjResult(interp, ret_list);
+
+	return(TCL_OK);
+}
+
+/* XXX:TODO: Merge this with list_certs */
+MODULE_SCOPE int tclpkcs11_list_keys(ClientData cd, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) {
+	struct tclpkcs11_interpdata *interpdata;
+	struct tclpkcs11_handle *handle;
+	Tcl_HashEntry *tcl_handle_entry;
+	Tcl_Obj *tcl_handle, *tcl_slotid;
+	long slotid_long;
+	Tcl_Obj *obj_label, *obj_id, *obj_exponent, *obj_modulus;
+	Tcl_Obj *ret_list, *curr_item_list;
+	int tcl_rv;
+
+	CK_SLOT_ID slotid;
+	CK_OBJECT_HANDLE hObject;
+	CK_ULONG ulObjectCount;
+	CK_ATTRIBUTE template[] = {
+	                           {CKA_CLASS, NULL, 0},
+	                           {CKA_LABEL, NULL, 0},
+	                           {CKA_ID, NULL, 0},
+	                           {CKA_KEY_TYPE, NULL, 0},
+	                           {CKA_PUBLIC_EXPONENT, NULL, 0},
+	                           {CKA_MODULUS, NULL, 0}
+	};
+	CK_ATTRIBUTE *curr_attr;
+	CK_ULONG curr_attr_idx;
+	CK_OBJECT_CLASS objectclass, *objectclass_p;
+	CK_KEY_TYPE keytype;
+	CK_RV chk_rv;
+
+	if (!cd) {
+		Tcl_SetObjResult(interp, Tcl_NewStringObj("invalid clientdata", -1));
+
+		return(TCL_ERROR);
+	}
+
+	if (objc != 3) {
+		Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \"pki::pkcs11::listkeys handle slot\"", -1));
+
+		return(TCL_ERROR);
+	}
+
+	tcl_handle = objv[1];
+	tcl_slotid = objv[2];
+
+	interpdata = (struct tclpkcs11_interpdata *) cd;
+
+	tcl_handle_entry = Tcl_FindHashEntry(&interpdata->handles, (const char *) tcl_handle);
+	if (!tcl_handle_entry) {
+		Tcl_SetObjResult(interp, Tcl_NewStringObj("invalid handle", -1));
+
+		return(TCL_ERROR);
+	}
+
+	handle = (struct tclpkcs11_handle *) Tcl_GetHashValue(tcl_handle_entry);
+	if (!handle) {
+		Tcl_SetObjResult(interp, Tcl_NewStringObj("invalid handle", -1));
+
+		return(TCL_ERROR);
+	}
+
+	tcl_rv = Tcl_GetLongFromObj(interp, tcl_slotid, &slotid_long);
+	if (tcl_rv != TCL_OK) {
+		return(tcl_rv);
+	}
+
+	slotid = slotid_long;
+
+	chk_rv = tclpkcs11_start_session(handle, slotid);
+	if (chk_rv != CKR_OK) {
+		Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
+
+		return(TCL_ERROR);
+	}
+
+	chk_rv = handle->pkcs11->C_FindObjectsInit(handle->session, NULL, 0);
+	if (chk_rv != CKR_OK) {
+		Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
+
+		return(TCL_ERROR);
+	}
+
 	ret_list = Tcl_NewObj();
 	while (1) {
 		chk_rv = handle->pkcs11->C_FindObjects(handle->session, &hObject, 1, &ulObjectCount);
@@ -1124,11 +1393,13 @@ MODULE_SCOPE int tclpkcs11_list_certs(ClientData cd, Tcl_Interp *interp, int obj
 			return(TCL_ERROR);
 		}
 
-		/* Extract certificate data */
+		/* Extract key data */
 		obj_label = NULL;
 		obj_id = NULL;
-		obj_cert = NULL;
-		objectclass = NULL;
+		objectclass_p = NULL;
+		obj_exponent = NULL;
+		obj_modulus = NULL;
+		keytype = -1;
 		for (curr_attr_idx = 0; curr_attr_idx < (sizeof(template) / sizeof(template[0])); curr_attr_idx++) {
 			curr_attr = &template[curr_attr_idx];
 
@@ -1136,14 +1407,17 @@ MODULE_SCOPE int tclpkcs11_list_certs(ClientData cd, Tcl_Interp *interp, int obj
 				continue;
 			}
 
+			if (curr_attr->type != CKA_CLASS && !objectclass) {
+				ckfree(curr_attr->pValue);
+				curr_attr->pValue = NULL;
+
+				continue;
+			}
+
 			switch (curr_attr->type) {
 				case CKA_CLASS:
-					objectclass = (CK_OBJECT_CLASS *) curr_attr->pValue;
-
-					if (*objectclass != CKO_CERTIFICATE) {
-						continue;
-					}
-
+					objectclass_p = (CK_OBJECT_CLASS *) curr_attr->pValue;
+					objectclass = *objectclass_p;
 					break;
 				case CKA_LABEL:
 					obj_label = Tcl_NewStringObj(curr_attr->pValue, curr_attr->ulValueLen);
@@ -1153,13 +1427,16 @@ MODULE_SCOPE int tclpkcs11_list_certs(ClientData cd, Tcl_Interp *interp, int obj
 					obj_id = tclpkcs11_bytearray_to_string(curr_attr->pValue, curr_attr->ulValueLen);
 
 					break;
-				case CKA_VALUE:
-					if (!objectclass) {
-						break;
-					}
-
-					obj_cert = Tcl_NewByteArrayObj(curr_attr->pValue, curr_attr->ulValueLen);
+				case CKA_KEY_TYPE:
+					/* Convert the ID into a readable string */
+					keytype = *((CK_KEY_TYPE *) curr_attr->pValue);
 
+					break;
+				case CKA_MODULUS:
+					obj_modulus = Tcl_NewByteArrayObj(curr_attr->pValue, curr_attr->ulValueLen);
+					break;
+				case CKA_PUBLIC_EXPONENT:
+					obj_exponent = Tcl_NewByteArrayObj(curr_attr->pValue, curr_attr->ulValueLen);
 					break;
 			}
 
@@ -1168,7 +1445,15 @@ MODULE_SCOPE int tclpkcs11_list_certs(ClientData cd, Tcl_Interp *interp, int obj
 		}
 
 		/* Add this certificate data to return list, if all found */
-		if (obj_label == NULL || obj_id == NULL || obj_cert == NULL) {
+		if (obj_label == NULL || obj_id == NULL || obj_exponent == NULL || obj_modulus == NULL || keytype == -1) {
+			continue;
+		}
+
+		if (objectclass != CKO_PUBLIC_KEY) {
+			continue;
+		}
+
+		if (keytype != CKK_RSA) {
 			continue;
 		}
 
@@ -1186,18 +1471,18 @@ MODULE_SCOPE int tclpkcs11_list_certs(ClientData cd, Tcl_Interp *interp, int obj
 		Tcl_ListObjAppendElement(interp, curr_item_list, Tcl_NewStringObj("pkcs11_label", -1));
 		Tcl_ListObjAppendElement(interp, curr_item_list, obj_label);
 
-		/* Call "::pki::x509::parse_cert" to parse the cert */
-		parse_cert_cmd = Tcl_NewObj();
-		Tcl_ListObjAppendElement(interp, parse_cert_cmd, Tcl_NewStringObj("::pki::x509::parse_cert", -1));
-		Tcl_ListObjAppendElement(interp, parse_cert_cmd, obj_cert);
+		/* Add the RSA key to the list */
+		Tcl_ListObjAppendElement(interp, curr_item_list, Tcl_NewStringObj("n", -1));
+		Tcl_ListObjAppendElement(interp, curr_item_list, Tcl_ObjPrintf("0x%s", Tcl_GetString(tclpkcs11_bytearray_to_string_from_obj(obj_modulus))));
 
-		tcl_rv = Tcl_EvalObjEx(interp, parse_cert_cmd, 0);
-		if (tcl_rv != TCL_OK) {
-			continue;
-		}
+		Tcl_ListObjAppendElement(interp, curr_item_list, Tcl_NewStringObj("e", -1));
+		Tcl_ListObjAppendElement(interp, curr_item_list, Tcl_ObjPrintf("0x%s", Tcl_GetString(tclpkcs11_bytearray_to_string_from_obj(obj_exponent))));
 
-		/* Add results of [parse_cert] to our return value */
-		Tcl_ListObjAppendList(interp, curr_item_list, Tcl_GetObjResult(interp));
+		Tcl_ListObjAppendElement(interp, curr_item_list, Tcl_NewStringObj("l", -1));
+		Tcl_ListObjAppendElement(interp, curr_item_list, Tcl_ObjPrintf("%i", Tcl_GetCharLength(obj_modulus) * 8));
+
+		Tcl_ListObjAppendElement(interp, curr_item_list, Tcl_NewStringObj("type", -1));
+		Tcl_ListObjAppendElement(interp, curr_item_list, Tcl_NewStringObj("rsa", -1));
 
 		/*
 		 * Override the "type" so that [array set] returns our new
@@ -1560,6 +1845,8 @@ MODULE_SCOPE int tclpkcs11_perform_pki(int encrypt, ClientData cd, Tcl_Interp *i
 
 	chk_rv = handle->pkcs11->C_FindObjectsInit(handle->session, template, sizeof(template) / sizeof(template[0]));
 	if (chk_rv != CKR_OK) {
+		tclpkcs11_close_session(handle);
+
 		Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
 
 		return(TCL_ERROR);
@@ -1568,10 +1855,12 @@ MODULE_SCOPE int tclpkcs11_perform_pki(int encrypt, ClientData cd, Tcl_Interp *i
 	foundObjs = 0;
 	chk_rv = handle->pkcs11->C_FindObjects(handle->session, &hObject, 1, &foundObjs);
 	if (chk_rv != CKR_OK) {
-		Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
-
 		handle->pkcs11->C_FindObjectsFinal(handle->session);
 
+		tclpkcs11_close_session(handle);
+
+		Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
+
 		return(TCL_ERROR);
 	}
 
@@ -1579,6 +1868,8 @@ MODULE_SCOPE int tclpkcs11_perform_pki(int encrypt, ClientData cd, Tcl_Interp *i
 	handle->pkcs11->C_FindObjectsFinal(handle->session);
 
 	if (foundObjs < 1) {
+		tclpkcs11_close_session(handle);
+
 		Tcl_SetObjResult(interp, Tcl_NewStringObj("PKCS11_ERROR MAYBE_LOGIN", -1));
 
 		return(TCL_ERROR);
@@ -1594,6 +1885,8 @@ MODULE_SCOPE int tclpkcs11_perform_pki(int encrypt, ClientData cd, Tcl_Interp *i
 			chk_rv = handle->pkcs11->C_SignInit(handle->session, &mechanism, hObject);
 		}
 		if (chk_rv != CKR_OK) {
+			tclpkcs11_close_session(handle);
+
 			Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
 
 			return(TCL_ERROR);
@@ -1641,6 +1934,8 @@ MODULE_SCOPE int tclpkcs11_perform_pki(int encrypt, ClientData cd, Tcl_Interp *i
 		}
 
 		if (chk_rv != CKR_OK) {
+			tclpkcs11_close_session(handle);
+
 			Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
 
 			return(TCL_ERROR);
@@ -1648,6 +1943,8 @@ MODULE_SCOPE int tclpkcs11_perform_pki(int encrypt, ClientData cd, Tcl_Interp *i
 	} else {
 		chk_rv = handle->pkcs11->C_DecryptInit(handle->session, &mechanism, hObject);
 		if (chk_rv != CKR_OK) {
+			tclpkcs11_close_session(handle);
+
 			Tcl_SetObjResult(interp, tclpkcs11_pkcs11_error(chk_rv));
 
 			return(TCL_ERROR);
@@ -1656,6 +1953,8 @@ MODULE_SCOPE int tclpkcs11_perform_pki(int encrypt, ClientData cd, Tcl_Interp *i
 		resultbuf_len = sizeof(resultbuf);
 		chk_rv = handle->pkcs11->C_Decrypt(handle->session, input, input_len, resultbuf, &resultbuf_len);
 		if (chk_rv != CKR_OK) {
+			tclpkcs11_close_session(handle);
+
 			if (chk_rv == CKR_BUFFER_TOO_SMALL) {
 				/* Terminate decryption operation */
 				handle->pkcs11->C_DecryptFinal(handle->session, NULL, 0);
@@ -1667,6 +1966,8 @@ MODULE_SCOPE int tclpkcs11_perform_pki(int encrypt, ClientData cd, Tcl_Interp *i
 		}
 	}
 
+	tclpkcs11_close_session(handle);
+
 	tcl_result = Tcl_NewByteArrayObj(resultbuf, resultbuf_len);
 
 	Tcl_SetObjResult(interp, tcl_result);
@@ -1851,6 +2152,11 @@ int Tclpkcs11_Init(Tcl_Interp *interp) {
 		return(TCL_ERROR);
 	}
 
+	tclCreatComm_ret = Tcl_CreateObjCommand(interp, "pki::pkcs11::listkeys", tclpkcs11_list_keys, interpdata, NULL);
+	if (!tclCreatComm_ret) {
+		return(TCL_ERROR);
+	}
+
 	tclCreatComm_ret = Tcl_CreateObjCommand(interp, "pki::pkcs11::login", tclpkcs11_login, interpdata, NULL);
 	if (!tclCreatComm_ret) {
 		return(TCL_ERROR);
