@@ -4,6 +4,14 @@ lappend auto_path [file join [pwd] work lib]
 lappend auto_path [file join [pwd] lib]
 
 set pkcs11_module "/usr/local/lib/libcackey.so"
+set pkcs11_module /home/rkeene/devel/cackey/libcackey.so
+set pkcs11_module /home/rkeene/devel/saml-idp/archive/gcp-pkcs11.so
+
+if {0} {
+	set env(PKCS11SPY) $pkcs11_module
+	set env(PKCS11SPY_OUTPUT) /dev/stderr
+	set pkcs11_module /usr/lib/x86_64-linux-gnu/pkcs11/pkcs11-spy.so
+}
 
 load ./tclpkcs11.so Tclpkcs11
 
@@ -19,8 +27,10 @@ foreach slotinfo $slots {
 	set slotflags [lindex $slotinfo 2]
 
 	if {[lsearch -exact $slotflags TOKEN_PRESENT] != -1} {
-		set token_slotlabel $slotlabel
-		set token_slotid $slotid
+		if {![info exists token_slotid]} {
+			set token_slotlabel $slotlabel
+			set token_slotid $slotid
+		}
 	}
 }
 
@@ -30,6 +40,21 @@ if {![info exists token_slotid]} {
 	exit 1
 }
 
+set pubKeys [pki::pkcs11::listkeys $handle $token_slotid]
+puts "Found [llength $pubKeys] keys"
+
+set orig "TestMsg"
+foreach keyinfo_list $pubKeys {
+	unset -nocomplain keyinfo
+	array set keyinfo $keyinfo_list
+	puts "Key: $keyinfo(pkcs11_label)"
+
+	set signature [pki::sign $orig $keyinfo_list sha256]
+	set verify [pki::verify $signature $orig $keyinfo_list]
+
+	puts "Signature valid: $verify"
+}
+
 set certs [pki::pkcs11::listcerts $handle $token_slotid]
 puts "Found [llength $certs] certificates"
 
@@ -77,6 +102,8 @@ foreach certinfo_list $certs {
 
 		break
 	}
+
+	puts "OK"
 }
 
 pki::pkcs11::unloadmodule $handle
